@@ -13,8 +13,9 @@ import textwrap
 from contextlib import nullcontext
 from itertools import dropwhile
 
-from funcparserlib.parser import forward_decl, many, maybe, oneplus, some
+from funcparserlib.parser import finished, forward_decl, many, maybe, oneplus, some
 
+from hy._compat import PY3_11
 from hy.compiler import Result, asty, hy_eval, mkexpr
 from hy.errors import HyEvalError, HyInternalError, HyTypeError
 from hy.macros import pattern_macro, require, require_reader
@@ -53,6 +54,7 @@ from hy.models import (
     String,
     Symbol,
     Tuple,
+    as_model,
     is_unpack,
 )
 from hy.reader import mangle, unmangle
@@ -73,6 +75,10 @@ def maybe_annotated(target):
     return pexpr(sym("annotate") + target + FORM) | target >> (lambda x: (x, None))
 
 
+def dotted(name):
+    return Expression(map(Symbol, [".", *name.split(".")]))
+
+
 # ------------------------------------------------
 # * Fundamentals
 # ------------------------------------------------
@@ -83,12 +89,12 @@ def compile_do(self, expr, root, body):
     return self._compile_branch(body)
 
 
-@pattern_macro(["eval-and-compile", "eval-when-compile"], [many(FORM)])
-def compile_eval_and_compile(compiler, expr, root, body):
+@pattern_macro(["eval-and-compile", "eval-when-compile", "do-mac"], [many(FORM)])
+def compile_eval_foo_compile(compiler, expr, root, body):
     new_expr = Expression([Symbol("do").replace(expr[0])]).replace(expr)
 
     try:
-        hy_eval(
+        value = hy_eval(
             new_expr + body,
             compiler.module.__dict__,
             compiler.module,
@@ -110,7 +116,9 @@ def compile_eval_and_compile(compiler, expr, root, body):
         raise HyEvalError(str(e), compiler.filename, body, compiler.source)
 
     return (
-        compiler._compile_branch(body)
+        compiler.compile(as_model(value))
+        if mangle(root) == "do_mac"
+        else compiler._compile_branch(body)
         if mangle(root) == "eval_and_compile"
         else Result()
     )
@@ -180,8 +188,7 @@ def render_quoted_form(compiler, form, level):
     elif op in ("unquote", "unquote-splice"):
         level -= 1
 
-    hytype = form.__class__
-    name = ".".join((hytype.__module__, hytype.__name__))
+    name = form.__class__.__name__
     body = [form]
 
     if isinstance(form, Sequence):
@@ -213,8 +220,7 @@ def render_quoted_form(compiler, form, level):
         if form.brackets is not None:
             body.extend([Keyword("brackets"), form.brackets])
 
-    ret = Expression([Symbol(name), *body]).replace(form)
-    return ret, False
+    return (Expression([dotted("hy.models." + name), *body]).replace(form), False)
 
 
 # ------------------------------------------------
@@ -222,9 +228,9 @@ def render_quoted_form(compiler, form, level):
 # ------------------------------------------------
 
 
-@pattern_macro(["not", "~"], [FORM], shadow=True)
+@pattern_macro(["not", "bnot"], [FORM], shadow=True)
 def compile_unary_operator(compiler, expr, root, arg):
-    ops = {"not": ast.Not, "~": ast.Invert}
+    ops = {"not": ast.Not, "bnot": ast.Invert}
     operand = compiler.compile(arg)
     return operand + asty.UnaryOp(expr, op=ops[root](), operand=operand.force_expr)
 
@@ -446,7 +452,7 @@ def compile_assign(
 
     ld_name = compiler.compile(name)
 
-    if result.temp_variables and isinstance(name, Symbol) and "." not in name:
+    if result.temp_variables and isinstance(name, Symbol):
         result.rename(compiler, compiler._nonconst(name))
         if not is_assignment_expr:
             # Throw away .expr to ensure that (setv ...) returns None.
@@ -698,7 +704,9 @@ loopers = many(
     ["for"], [brackets(loopers), many(notpexpr("else")) + maybe(dolike("else"))]
 )
 @pattern_macro(["lfor", "sfor", "gfor"], [loopers, FORM])
-@pattern_macro(["dfor"], [loopers, brackets(FORM, FORM)])
+@pattern_macro(["dfor"], [loopers, finished])
+# Here `finished` is a hack replacement for FORM + FORM:
+# https://github.com/vlasovskikh/funcparserlib/issues/75
 def compile_comprehension(compiler, expr, root, parts, final):
     node_class = {
         "for": asty.For,
@@ -715,6 +723,13 @@ def compile_comprehension(compiler, expr, root, parts, final):
         # Compile the parts.
         if is_for:
             parts = parts[0]
+        if node_class is asty.DictComp:
+            if not (parts and parts[-1].tag == "for"):
+                raise compiler._syntax_error(
+                    parts[-1] if parts else parts,
+                    "`dfor` must end with key and value forms",
+                )
+            final = parts.pop().value
         if not parts:
             return Result(
                 expr=asty.parse(
@@ -1069,7 +1084,8 @@ _pattern.define(
         | pexpr(keepsym("|"), many(_pattern))
         | braces(many(LITERAL + _pattern), maybe(pvalue("unpack-mapping", SYM)))
         | pexpr(
-            notsym(".", "|", "unpack-mapping", "unpack-iterable"),
+            pexpr(keepsym("."), oneplus(SYM))
+            | notsym(".", "|", "unpack-mapping", "unpack-iterable"),
             many(parse_if(lambda x: not isinstance(x, Keyword), _pattern)),
             many(KEYWORD + _pattern),
         )
@@ -1160,11 +1176,7 @@ def compile_pattern(compiler, pattern):
             value,
             value=compiler.compile(value).force_expr.value,
         )
-    elif (
-        isinstance(value, (String, Integer, Float, Complex, Bytes))
-        or isinstance(value, Symbol)
-        and "." in value
-    ):
+    elif isinstance(value, (String, Integer, Float, Complex, Bytes)):
         return asty.MatchValue(
             value,
             value=compiler.compile(value).expr,
@@ -1208,11 +1220,15 @@ def compile_pattern(compiler, pattern):
             )
         )
     elif isinstance(value, Expression):
-        root, args, kwargs = value
+        head, args, kwargs = value
         keywords, values = zip(*kwargs) if kwargs else ([], [])
         return asty.MatchClass(
             value,
-            cls=compiler.scope.access(asty.Name(root, id=mangle(root), ctx=ast.Load())),
+            cls=compiler.compile(
+              # `head` could be a symbol or a dotted form.
+                (head[:1] + head[1]).replace(head)
+                if type(head) is Expression
+                else head).expr,
             patterns=[compile_pattern(compiler, v) for v in args],
             kwd_attrs=[kwd.name for kwd in keywords],
             kwd_patterns=[compile_pattern(compiler, value) for value in values],
@@ -1220,7 +1236,7 @@ def compile_pattern(compiler, pattern):
     elif isinstance(value, Keyword):
         return asty.MatchClass(
             value,
-            cls=compiler.compile(Symbol("hy.models.Keyword")).expr,
+            cls=compiler.compile(dotted("hy.models.Keyword")).expr,
             patterns=[
                 asty.MatchValue(value, value=asty.Constant(value, value=value.name))
             ],
@@ -1258,10 +1274,10 @@ def compile_raise_expression(compiler, expr, root, exc, cause):
 @pattern_macro(
     "try",
     [
-        many(notpexpr("except", "else", "finally")),
+        many(notpexpr("except", "except*", "else", "finally")),
         many(
             pexpr(
-                sym("except"),
+                keepsym("except") | keepsym("except*"),
                 brackets() | brackets(FORM) | brackets(SYM, FORM),
                 many(FORM),
             )
@@ -1277,9 +1293,57 @@ def compile_try_expression(compiler, expr, root, body, catchers, orelse, finalbo
 
     handler_results = Result()
     handlers = []
+    except_syms_seen = set()
     for catcher in catchers:
-        handler_results += compile_catch_expression(
-            compiler, catcher, return_var, *catcher
+        # exceptions catch should be either:
+        # [[list of exceptions]]
+        # or
+        # [variable [list of exceptions]]
+        # or
+        # [variable exception]
+        # or
+        # [exception]
+        # or
+        # []
+        except_sym, exceptions, ebody = catcher
+        if not PY3_11 and except_sym == Symbol("except*"):
+            hy_compiler._syntax_error(except_sym, "`{}` requires Python 3.11 or later")
+        except_syms_seen.add(str(except_sym))
+        if len(except_syms_seen) > 1:
+            raise compiler._syntax_error(
+                except_sym, "cannot have both `except` and `except*` on the same `try`"
+            )
+
+        name = None
+        if len(exceptions) == 2:
+            name = mangle(compiler._nonconst(exceptions[0]))
+
+        exceptions_list = exceptions[-1] if exceptions else List()
+        if isinstance(exceptions_list, List):
+            if len(exceptions_list):
+                # [FooBar BarFoo] → catch Foobar and BarFoo exceptions
+                elts, types, _ = compiler._compile_collect(exceptions_list)
+                types += asty.Tuple(exceptions_list, elts=elts, ctx=ast.Load())
+            else:
+                # [] → all exceptions caught
+                types = Result()
+        else:
+            types = compiler.compile(exceptions_list)
+
+        # Create a "fake" scope for the exception variable.
+        # See: https://docs.python.org/3/reference/compound_stmts.html#the-try-statement
+        with compiler.scope.create(ScopeLet) as scope:
+            if name:
+                scope.add(name, name)
+            ebody = compiler._compile_branch(ebody)
+        ebody += asty.Assign(catcher, targets=[return_var], value=ebody.force_expr)
+        ebody += ebody.expr_as_stmt()
+
+        handler_results += types + asty.ExceptHandler(
+            catcher,
+            type=types.expr,
+            name=name,
+            body=ebody.stmts or [asty.Pass(catcher)],
         )
         handlers.append(handler_results.stmts.pop())
 
@@ -1318,50 +1382,10 @@ def compile_try_expression(compiler, expr, root, body, catchers, orelse, finalbo
     )
     body = body.stmts or [asty.Pass(expr)]
 
-    x = asty.Try(expr, body=body, handlers=handlers, orelse=orelse, finalbody=finalbody)
-    return handler_results + x + returnable
-
-
-def compile_catch_expression(compiler, expr, var, exceptions, body):
-    # exceptions catch should be either:
-    # [[list of exceptions]]
-    # or
-    # [variable [list of exceptions]]
-    # or
-    # [variable exception]
-    # or
-    # [exception]
-    # or
-    # []
-
-    name = None
-    if len(exceptions) == 2:
-        name = mangle(compiler._nonconst(exceptions[0]))
-
-    exceptions_list = exceptions[-1] if exceptions else List()
-    if isinstance(exceptions_list, List):
-        if len(exceptions_list):
-            # [FooBar BarFoo] → catch Foobar and BarFoo exceptions
-            elts, types, _ = compiler._compile_collect(exceptions_list)
-            types += asty.Tuple(exceptions_list, elts=elts, ctx=ast.Load())
-        else:
-            # [] → all exceptions caught
-            types = Result()
-    else:
-        types = compiler.compile(exceptions_list)
-
-    # Create a "fake" scope for the exception variable.
-    # See: https://docs.python.org/3/reference/compound_stmts.html#the-try-statement
-    with compiler.scope.create(ScopeLet) as scope:
-        if name:
-            scope.add(name, name)
-        body = compiler._compile_branch(body)
-    body += asty.Assign(expr, targets=[var], value=body.force_expr)
-    body += body.expr_as_stmt()
-
-    return types + asty.ExceptHandler(
-        expr, type=types.expr, name=name, body=body.stmts or [asty.Pass(expr)]
+    x = (asty.TryStar if "except*" in except_syms_seen else asty.Try)(
+        expr, body=body, handlers=handlers, orelse=orelse, finalbody=finalbody
     )
+    return handler_results + x + returnable
 
 
 # ------------------------------------------------
@@ -1458,9 +1482,6 @@ def compile_function_node(compiler, expr, node, decorators, name, args, returns,
     ],
 )
 def compile_macro_def(compiler, expr, root, name, params, body):
-    if "." in name:
-        raise compiler._syntax_error(name, "periods are not allowed in macro names")
-
     ret = Result() + compiler.compile(
         Expression(
             [
@@ -1469,7 +1490,7 @@ def compile_macro_def(compiler, expr, root, name, params, body):
                     [
                         Expression(
                             [
-                                Symbol("hy.macros.macro"),
+                                dotted("hy.macros.macro"),
                                 str(name),
                             ]
                         ),
@@ -1662,28 +1683,40 @@ def compile_class_expression(compiler, expr, root, decorators, name, rest):
 # * `import` and `require`
 # ------------------------------------------------
 
+module_name_pattern = SYM | pexpr(
+    some(lambda x: isinstance(x, Symbol) and not str(x[0]).strip(".")) + oneplus(SYM)
+)
 
-def importlike(*name_types):
-    name = some(lambda x: isinstance(x, name_types) and "." not in x)
+
+def module_name_str(x):
     return (
-        keepsym("*")
-        | (keepsym(":as") + name)
-        | brackets(many(name + maybe(sym(":as") + name)))
+        ".".join(map(mangle, x[1][x[1][0] == Symbol("None") :]))
+        if isinstance(x, Expression)
+        else str(x)
+        if isinstance(x, Symbol) and not x.strip(".")
+        else mangle(x)
     )
 
 
+importlike = (
+    keepsym("*")
+    | (keepsym(":as") + SYM)
+    | brackets(many(SYM + maybe(sym(":as") + SYM)))
+)
+
+
 def assignment_shape(module, rest):
     prefix = ""
     assignments = "EXPORTS"
     if rest is None:
         # (import foo)
-        prefix = module
+        prefix = module_name_str(module)
     elif rest == Symbol("*"):
         # (import foo *)
         pass
     elif rest[0] == Keyword("as"):
         # (import foo :as bar)
-        prefix = rest[1]
+        prefix = mangle(rest[1])
     else:
         # (import foo [bar baz :as MyBaz bing])
         assignments = [(k, v or k) for k, v in rest[0]]
@@ -1694,11 +1727,11 @@ def assignment_shape(module, rest):
     "require",
     [
         many(
-            SYM
+            module_name_pattern
             + times(
                 0,
                 2,
-                (maybe(sym(":macros")) + importlike(Symbol))
+                (maybe(sym(":macros")) + importlike)
                 | (keepsym(":readers") + (keepsym("*") | brackets(many(SYM)))),
             )
         )
@@ -1722,13 +1755,16 @@ def compile_require(compiler, expr, root, entries):
         readers = readers and readers[0]
 
         prefix, assignments = assignment_shape(module, rest)
-        ast_module = mangle(module)
+        module_name = module_name_str(module)
+        if isinstance(module, Expression) and module[1][0] == Symbol("None"):
+            # Prepend leading dots to `module_name`.
+            module_name = str(module[0]) + module_name
 
         # we don't want to import all macros as prefixed if we're specifically
         # importing readers but not macros
         # (require a-module :readers ["!"])
         if (rest or not readers) and require(
-            ast_module, compiler.module, assignments=assignments, prefix=prefix
+            module_name, compiler.module, assignments=assignments, prefix=prefix
         ):
             # Actually calling `require` is necessary for macro expansions
             # occurring during compilation.
@@ -1737,8 +1773,8 @@ def compile_require(compiler, expr, root, entries):
             ret += compiler.compile(
                 Expression(
                     [
-                        Symbol("hy.macros.require"),
-                        String(ast_module),
+                        dotted("hy.macros.require"),
+                        String(module_name),
                         Symbol("None"),
                         Keyword("assignments"),
                         (
@@ -1759,22 +1795,22 @@ def compile_require(compiler, expr, root, entries):
                 if readers == Symbol("*")
                 else ["#" + reader for reader in readers[0]]
             )
-            if require_reader(ast_module, compiler.module, reader_assignments):
+            if require_reader(module_name, compiler.module, reader_assignments):
                 ret += compiler.compile(
                     mkexpr(
                         "do",
                         mkexpr(
-                            "hy.macros.require-reader",
-                            String(ast_module),
+                            dotted("hy.macros.require-reader"),
+                            String(module_name),
                             "None",
                             [reader_assignments],
                         ),
                         mkexpr(
                             "eval-when-compile",
                             mkexpr(
-                                "hy.macros.enable-readers",
+                                dotted("hy.macros.enable-readers"),
                                 "None",
-                                "hy.&reader",
+                                dotted("hy.&reader"),
                                 [reader_assignments],
                             ),
                         ),
@@ -1784,31 +1820,25 @@ def compile_require(compiler, expr, root, entries):
     return ret
 
 
-@pattern_macro("import", [many(SYM + maybe(importlike(Symbol)))])
+@pattern_macro("import", [many(module_name_pattern + maybe(importlike))])
 def compile_import(compiler, expr, root, entries):
     ret = Result()
 
     for entry in entries:
-        assignments = "EXPORTS"
-        prefix = ""
-
         module, _ = entry
         prefix, assignments = assignment_shape(*entry)
-        ast_module = mangle(module)
-
-        module_name = ast_module.lstrip(".")
-        level = len(ast_module) - len(module_name)
+        module_name = module_name_str(module)
         if assignments == "EXPORTS" and prefix == "":
             node = asty.ImportFrom
             names = [asty.alias(module, name="*", asname=None)]
         elif assignments == "EXPORTS":
-            compiler.scope.define(mangle(prefix))
+            compiler.scope.define(prefix)
             node = asty.Import
             names = [
                 asty.alias(
                     module,
-                    name=ast_module,
-                    asname=mangle(prefix) if prefix != module else None,
+                    name=module_name,
+                    asname=prefix if prefix != module_name else None,
                 )
             ]
         else:
@@ -1821,7 +1851,18 @@ def compile_import(compiler, expr, root, entries):
                         module, name=mangle(k), asname=None if v == k else mangle(v)
                     )
                 )
-        ret += node(expr, module=module_name or None, names=names, level=level)
+        ret += node(
+            expr,
+            module=module_name if module_name and module_name.strip(".") else None,
+            names=names,
+            level=(
+                len(module[0])
+                if isinstance(module, Expression) and module[1][0] == Symbol("None")
+                else len(module)
+                if isinstance(module, Symbol) and not module.strip(".")
+                else 0
+            ),
+        )
 
     return ret
 
@@ -1833,27 +1874,33 @@ def compile_import(compiler, expr, root, entries):
 
 @pattern_macro("assert", [FORM, maybe(FORM)])
 def compile_assert_expression(compiler, expr, root, test, msg):
-    if msg is None or type(msg) is Symbol:
-        ret = compiler.compile(test)
-        return ret + asty.Assert(
-            expr,
-            test=ret.force_expr,
-            msg=(None if msg is None else compiler.compile(msg).force_expr),
-        )
+    test = compiler.compile(test)
+    if msg:
+        msg = compiler.compile(msg)
 
-    # The `msg` part may involve statements, which we only
-    # want to be executed if the assertion fails. Rewrite the
-    # form to set `msg` to a variable.
-    msg_var = compiler.get_anon_var()
-    return compiler.compile(
-        mkexpr(
-            "if",
-            mkexpr("and", "__debug__", mkexpr("not", [test])),
-            mkexpr(
-                "do", mkexpr("setv", msg_var, [msg]), mkexpr("assert", "False", msg_var)
-            ),
-            Symbol("None"),
-        ).replace(expr)
+    if not (test.stmts or (msg and msg.stmts)):
+        return asty.Assert(expr, test=test.force_expr, msg=msg and msg.force_expr)
+
+    return asty.If(
+        expr,
+        test=asty.Name(expr, id="__debug__", ctx=ast.Load()),
+        orelse=[],
+        body=test.stmts
+        + [
+            asty.If(
+                test,
+                test=asty.UnaryOp(test, op=ast.Not(), operand=test.force_expr),
+                orelse=[],
+                body=(msg.stmts if msg else [])
+                + [
+                    asty.Assert(
+                        expr,
+                        test=asty.Constant(test, value=False),
+                        msg=msg and msg.force_expr,
+                    )
+                ],
+            )
+        ],
     )
 
 
@@ -1871,7 +1918,12 @@ def compile_let(compiler, expr, root, bindings, body):
 
 
 @pattern_macro(
-    "unquote unquote-splice unpack-mapping except finally else".split(), [many(FORM)]
+    "pragma unquote unquote-splice unpack-mapping except except* finally else".split(),
+    [many(FORM)],
 )
 def compile_placeholder(compiler, expr, root, body):
-    raise ValueError(f"`{root}` is not allowed here")
+    raise ValueError(
+        "`{}` is not allowed {}".format(
+            root, "in this version of Hy" if root == "pragma" else "here"
+        )
+    )
